def yes_no(answer):
	yes = set(['y'])
	no = set(['n'])
	 
	while True:
		choice = input(answer).lower()
		if choice in yes:
		   return True
		elif choice in no:
		   return False
		else:
		   print("Please respond with 'y' or 'n' ")


import pandas as pd
import numpy as np
from sklearn import tree

from ..filters.Filter import Filter

class ImputationFilter(Filter):
	def applyFilter(self, dataFrame:pd.DataFrame):
		cols = dataFrame.columns 

		if len(cols) < 4:
			# print('DataFrame has less than 4 columns. Imputing missing values via interpolation.')
			for col in cols:
				values = dataFrame[col]

				# check if any nans exist in this column:
				if values.hasnans:
					# print('Found missing values in column '+ col)
					values.fillna(method = 'backfill', inplace = True)
					print('[Imputation-Filter] Added missing values in column '+ col + ' via backfilling.')

				dataFrame[col] = values

		else:
			# print('DataFrame has many columns. Imputing missing values via decision tree regression.')

			for col in cols: 
				values = dataFrame[col]

				# check if any nans exist in this column:
				if values.hasnans:
					# print('Found missing values in column '+ col)
					data = dataFrame

					# force conversion to numeric, else nan
					train_set = data.apply(lambda x: pd.to_numeric(x, errors='coerce')).dropna()
					train_val = train_set[col].values
					train_set = train_set.loc[:, data.columns != col]
					data = data.loc[:, data.columns != col]

					#find all indices in values that are nan and remove them from training data
					nan_indices = np.where(values.isnull())[0]

					# train decision tree regressor
					dTree = tree.DecisionTreeRegressor()

					dTree.fit(train_set, train_val)

					# infer missing values
					for i in nan_indices:
						try:
							values[i] = dTree.predict(data.iloc[i].values.reshape(1,-1))[0]
						except:
							pass 
					print('[Imputation-Filter]: Added missing values in column '+ col + ' via decision tree regression.')


				dataFrame[col] = values

				# check if any nans exist in this column:
				if values.hasnans:
					# print('Found unhandled missing values in column '+ col)
					values.fillna(method = 'backfill', inplace = True)
					print('[Imputation-Filter]: Added missing values in column '+ col + ' via backfilling.')
					dataFrame[col] = values


		return dataFrame


import pandas as pd

from ..filters.Filter import Filter


class FormatDataFrame(Filter):

    df: pd.DataFrame
    names = pd.DataFrame()
    units: pd.DataFrame

    def fixRowFormat(self):
        
        if len((self.df.columns)) == 2:
            tempDF = self.df
            try:
                tempDF.pivot(columns=tempDF.columns[0], values=tempDF.columns[1])
                tempDF.rename(
                columns={tempDF.columns[0]: "TIMESTAMP"}, inplace=True, errors="raise")
                self.df = tempDF
                return True
            except:
                print("[formatData-Filter]: Transformation to matrix format failed!")
                return False
        else:
            print("[formatData-Filter]: Did not convert to matrix format!")
            return False

    def dropNonDataRows(self):
        
        tempDF = self.df
        if not pd.api.types.is_datetime64_dtype(tempDF.index):
            tempDF.rename(
                columns={tempDF.columns[0]: "TIMESTAMP"}, inplace=True, errors="raise")
            try:
                start = tempDF["TIMESTAMP"].where(tempDF["TIMESTAMP"].str.contains(
                    "time", case=False, na=False)).first_valid_index()+1
                self.names = tempDF.iloc[start-2, 1::]
                self.units = tempDF.iloc[start-1, 1::]
                for x in range(start):
                    tempDF = tempDF.drop([x])
                self.df = tempDF
                return True
            except:
                return False
        else:
            print("[formatData-Filter]: Timestamp is already in datetime format!")
            return False

    def setTimestampAsIndex(self):

        # Make first column datetime and set as index. Errors are set as NaT
        tempDF = self.df
        try: 
            if not pd.api.types.is_datetime64_dtype(tempDF.index):
                tempDF["TIMESTAMP"] = pd.to_datetime(tempDF["TIMESTAMP"], infer_datetime_format=True, errors='coerce')
                tempDF = tempDF.set_index("TIMESTAMP")
                self.df = tempDF
                return True
            else:
                print("[formatData-Filter]: Timestamp is already in datetime format!")
                return False
        except:
            return False

    def setColumnNames(self):
        try:
            if not self.names.size == 0:
                self.df.rename(self.names, axis='columns', inplace=True)
                return True
            else:
                print("[formatData-Filter]: No names to set as column names found!")
                return False
        except:
            return False


    def cleanTypes(self):

        try: 
            tempDF = self.df
            tempDF.apply(pd.to_numeric, errors='coerce')
            self.df = tempDF
            return True
        except:
            return False

    def applyFilter(self, dataFrame: pd.DataFrame):
        
        self.df = dataFrame
        
        if self.fixRowFormat():
            print("[formatData-Filter]: Fixed Row Format")
        else:
            if self.dropNonDataRows():
                print("[formatData-Filter]: Dropped Non Data Rows")
            else:
                print("[formatData-Filter]: Dropping name columns failed!")

            if self.setColumnNames():
                print("[formatData-Filter]: Set Columnames")
            else:
                print("[formatData-Filter]: Setting column names failed!")

        if self.setTimestampAsIndex():
            print("[formatData-Filter]: Set Timestamp as Index")
        else:
            print("[formatData-Filter]: Setting timestamp as index failed!")

        if self.cleanTypes():
            print("[cleanTypes-Filter]: Casted the dataFrame to numeric")
        else:
            print("[formatData-Filter]: Casting the dataFrame to numeric failed!")

        return self.df



import pandas as pd

from .cleaners.FormatData import FormatDataFrame

from .preFilters.DuplicateFilter import DuplicateFilter
from .preFilters.ImputationFilter import ImputationFilter
from .preFilters.EquiDistanceFilter import EquiDistanceFilter
from .preFilters.FillTimestampsFilter import FillTimestampsFilter
from .preFilters.FixTimezoneFilter import FixTimezoneFilter

class PreFilterBuilder:


    dataFrame: pd.DataFrame


    def __init__(self, data=None):
        if isinstance(data, pd.DataFrame):
            self.dataFrame = self.formatData(data)
        elif isinstance(data, list):
            dataFrames = []
            for x in data:
                if isinstance(x, str):
                    self.dataFrame = self.formatData(pd.read_csv("dirty_data/" + x))
                    self.filterFixTimezone()
                    dataFrames.append(self.dataFrame)
                elif isinstance(x, pd.DataFrame):
                    dataFrames.append(self.formatData(x))
                else:
                    print("[PreFilterBuilder]: No Valid Input!")
            self.dataFrame = pd.concat([x for x in dataFrames ], axis=1)
        elif data is None:
            self.dataFrame = pd.DataFrame
        else:
            print("FilterBuilder: No Valid Input!")


    def getDataFrame(self):
        return self.dataFrame


    def formatData(self, df):
        filter = FormatDataFrame()
        dataFrame = filter.applyFilter(df)
        return dataFrame
        

    def filterEquidistance(self):
        filter = EquiDistanceFilter()
        self.dataFrame = filter.applyFilter(self.dataFrame)
        return self
    
    def filterFillTimestamps(self):
        filter = FillTimestampsFilter()
        self.dataFrame = filter.applyFilter(self.dataFrame)
        return self

    def filterFixTimezone(self):
        filter = FixTimezoneFilter()
        self.dataFrame = filter.applyFilter(self.dataFrame)
        return self
    
    def filterDuplicates(self):
        filter = DuplicateFilter()
        self.dataFrame = filter.applyFilter(self.dataFrame)
        return self

    def filterImputation(self):
        filter = ImputationFilter()
        self.dataFrame = filter.applyFilter(self.dataFrame)
        return self



import datetime
import pandas as pd

from tzlocal import get_localzone

from ..filters.Filter import Filter

class FixTimezoneFilter(Filter):

    def applyFilter(self, dataFrame:pd.DataFrame):
        # Check and set timezone information
        # print(dataFrame.head())
        if str(dataFrame.index.tzinfo) == "None":
            print("[FixTimezone-Filter]: No timezone information found. Setting local timezone: " + get_localzone().zone)
            dataFrame = dataFrame.tz_localize(tz = get_localzone())
        else:
            dataFrame = dataFrame.tz_convert(tz = get_localzone())       
        
        return dataFrame




import pandas as pd
import numpy as np
from dtwalign import dtw
from fastdtw import fastdtw

from .Filter import Filter

class LagFilter(Filter):
	def applyFilter(self, dataFrame:pd.DataFrame, columnPairs):
		# notes:
		# lag is defined as a time series x, which represents the time series y shifted by the lag z (expressed in time)
		# lagged time series can also be related time series (one is input, the other is output) --> extremely difficult to differentiate without knowledge
		# calculate lagged correlation (contrary to "classical" correlation) has the advantage of detecting lagged variables
		# different approaches for time lag correlation https://towardsdatascience.com/four-ways-to-quantify-synchrony-between-time-series-data-b99136c4a9c9
		# ideas: Pearson, Dynamic Time Warping, Partial Autocorrelation
		# to find all lagged variables, ones need to iterate through all variables and all timeshifts and compare them...
		# another approach is to find the maximum cross-correlation coefficient https://stackoverflow.com/questions/41492882/find-time-shift-of-two-signals-using-cross-correlation/56432463
		# see also: https://stats.stackexchange.com/questions/482927/estimate-the-delay-between-two-signal
		# but before using the Lag Filter, one should remove noise

		# calculate cross-correlation between columns
		# for i in len(dataFrame.columns)-2:
		# j = i+1
		# for j in len(dataFrame.columns)-1:
		
		# find lagged variables
		# where cross-correlation curve = autocorrelation curve? or where ac = ac?
		
		# use dynamic time warping to align lagged variables
		# tutorial: https://htmlpreview.github.io/?https://github.com/statefb/dtwalign/blob/master/example/example.html
		
		for columnPair in columnPairs:
			# print(columnPair[0])
			values_list_1 = dataFrame.iloc[0:100,columnPair[0]].to_numpy()
			values_list_2 = dataFrame.iloc[0:100,columnPair[1]].to_numpy()
			res = dtw(values_list_1, values_list_2)

			# warp both columns into "alignment path" --> warped column
			x_path = res.path[:,0]
			y_path = res.path[:,1]
			dataFrame.iloc[0:100,columnPair[0]] = dataFrame.iloc[x_path, columnPair[0]]
			dataFrame.iloc[0:100,columnPair[1]] = dataFrame.iloc[y_path, columnPair[1]]
			print("dtw distance: {}".format(res.normalized_distance))

		return dataFrame


import pandas as pd
import numpy as np
from sklearn.metrics.pairwise import cosine_similarity
from scipy.spatial import distance_matrix

from ..filters.Filter import Filter

class DuplicateFilter(Filter):
	def applyFilter(self, dataFrame:pd.DataFrame):
		# EASY METHOD for detecting exact duplicates
		# iterate through all columns in dataFrame and find equal columns (duplicates) + delete duplicate columns
		listOfColumnPairs_exact = []
		dataFrame_new = dataFrame
		# print(dataFrame_new.shape)
		for i in range(0, len(dataFrame.columns)-2):
			for j in range(i + 1, len(dataFrame.columns)-1):
				if dataFrame.iloc[:,i].equals(dataFrame.iloc[:,j]):
					listOfColumnPair_exact = [i, j]
					listOfColumnPairs_exact.append(listOfColumnPair_exact)
					# rename column i with new name ("i, j"), if column header are different
					if dataFrame_new.columns[listOfColumnPair_exact[0]] == dataFrame_new.columns[listOfColumnPair_exact[1]]:
						new_columnheader = dataFrame_new.columns[listOfColumnPair_exact[0]]
					else:
						new_columnheader = str(dataFrame_new.columns[listOfColumnPair_exact[0]]) + ", " + str(dataFrame_new.columns[listOfColumnPair_exact[1]])
					dataFrame_new.rename(columns = {str(dataFrame_new.columns[listOfColumnPair_exact[0]]):str(new_columnheader)}, inplace=True)
					# delete duplicate column
					dataFrame_new.drop(dataFrame_new.columns[listOfColumnPair_exact[1]], inplace=True, axis=1)
				else:
					pass
		# print(listOfColumnPairs_exact)
		# print(dataFrame_new.shape)

		return dataFrame

		'''# ADVANCED METHOD for detecting different kinds of duplicates
        # calculate similarity of columns, using the cosine similirarity and euclidean distance
		# article https://medium.com/@sasi24/cosine-similarity-vs-euclidean-distance-e5d9a9375fc8 compares both similarity measures
		# both measures have their justification to be used
		similarity_matrix_cosine = 1-cosine_similarity(dataFrame)
		similarity_matrix_euclidean = distance_matrix(dataFrame.values, dataFrame.values)

		# define similarity value for exact, linearTransformed and partial duplicates
		# check if coordinates of exact_duplicates are also 0 in the euclidean distance
		# if YES = exact duplicates
		# if NO = linear transformed duplicates
		similarity_matrix = similarity_matrix_cosine + similarity_matrix_euclidean
		exact_duplicates = np.where(simularity_matrix = 0)
		linearTransformed_duplicates = np.where(simularity_matrix_cosine = 0 and similarity_matrix_euclidean != 0)
		partial_duplicates = np.where(similarity_matrix_cosine > 0.99 & similarity_matrix_cosine < 1)

		# zip the 2 arrays to get the column pairs
		listOfColumnPairs_exact = list(zip(exact_duplicates[0], exact_duplicates[1]))
		listOfColumnPairs_linearTransformed= list(zip(linearTransformed_duplicates[0], linearTransformed_duplicates[1]))
		listOfColumnPairs_partial = list(zip(partial_duplicates[0], partial_duplicates[1]))

		# iterate over the list of coordinates
		for cord_exact in listOfColumnPairs_exact:
			print(cord_exact)

		for cord_linearTransformed in listOfColumnPairs_linearTransformed:
			print(cord_linearTransformed)

		for cord_partial in listOfColumnPairs_partial:
			print(cord_partial)
			
		return dataFrame_new, listOfColumnPairs_exact, listOfColumnPairs_linearTransformed, listOfColumnPairs_partial'''



import pandas as pd

from .filters.Filter import Filter
from .filters.DecimalShiftFilter import DecimalShiftFilter
from .filters.NoiseFilter import NoiseFilter
from .filters.LagFilter import LagFilter

class FilterBuilder:
	dataFrame: pd.DataFrame

	def __init__(self, data=None):
		if isinstance(data, str):
			self.dataFrame = pd.read_csv(data)
		elif isinstance(data, pd.DataFrame):
			self.dataFrame = data
		elif data is None:
			self.dataFrame = pd.DataFrame
		else:
			print("FilterBuilder: No Valid Input!")

	def getDataFrame(self):
		return self.dataFrame

	def setDataFrame(self, df:pd.DataFrame):
		self.dataFrame = df
		return self

	def filterDecimalShift(self):
		filter = DecimalShiftFilter()
		self.dataFrame = filter.applyFilter(self.dataFrame)
		return self

	def filterNoise(self, columnlist):
		filter = NoiseFilter()
		self.dataFrame = filter.applyFilter(self.dataFrame, columnlist)
		return self.dataFrame
	
	def filterLag(self, columnPairs):
		filter = LagFilter()
		self.dataFrame = filter.applyFilter(self.dataFrame, columnPairs)
		return self.dataFrame



import pandas as pd
import numpy as np
from sklearn.svm import OneClassSVM

from .Filter import Filter

class DecimalShiftFilter(Filter):
	def applyFilter(self, dataFrame:pd.DataFrame):
		# print("UnitValueFilter. DataFrame: " + str(dataFrame))

		for col in dataFrame.columns:
			values = dataFrame.loc[:,col]

			try:
				values = pd.to_numeric(values)

				# calculate change factor between two timestamps
				change_factor = []
				change_factor.append(0)
				for k in range(1, len(values)):
					change_factor.append(values[k]/values[k-1])

				multiplier = [round(x, 2) if x < 1 else round(x, 1) for x in change_factor]
				multiplier_low = [i for i,x in enumerate(multiplier) if x==0.1]
				multiplier_high = [i for i,x in enumerate(multiplier) if x==10]

				for i in range(min(len(multiplier_high), len(multiplier_high))):
					if multiplier_low[i] < multiplier_high[i]:
						for n in range(multiplier_low[i], multiplier_high[i]):
							values[n] = values[n] * 10
						print('[Decimalshift-Filter]: Detected comma error in ' + col + '.')
					elif multiplier_low[i] > multiplier_high[i]:
						for n in range(multiplier_high[i], multiplier_low[i]):
							values[n] = values[n] / 10
						print('[Decimalshift-Filter]: Detected comma error in ' + col + '.')

				dataFrame[col] = values
				
			except Exception:
				pass

		return dataFrame


import pandas as pd
import numpy as np

from ..filters.Filter import Filter

class EquiDistanceFilter(Filter):
    def applyFilter(self, dataFrame:pd.DataFrame):
        # find most frequent index distance value
        # create two new columns for the index and the difference between two neighbouring index values
        dataFrame['tvalue'] = dataFrame.index
        dataFrame['index-delta'] = (dataFrame['tvalue']-dataFrame['tvalue'].shift())
        # calculate most frequenst distance value by using the mode of the column
        dataOffset = int(dataFrame['index-delta'].mode().dt.total_seconds())
        # drop the created columns
        dataFrame.drop(columns=['tvalue', 'index-delta'])
        # create the offset (as a string) in seconds for the resample() method
        resample_dataOffset = str(dataOffset) + 'S'
        # use the resample() method to create equidistance
        dataFrame = dataFrame.resample(resample_dataOffset).sum()

        return dataFrame


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
  name = 'SciFiPi',
  version = '0.0.4', 				
  license='MIT',
  description = 'SciFiPi - The scientific filtering pipeline to clean your machine learning datasets',
  long_description = long_description,
  long_description_content_type="text/markdown",
  author = 'HSU Institute of Automation',
  #author_email = 'your.email@domain.com',
  url = 'https://github.com/GianMewes/KEEN',
  download_url = 'https://github.com/GianMewes/KEEN/archive/refs/tags/v0.0.4.tar.gz',
  keywords = ['Data Preprocessing', 'Data Filtering', 'Machine Learning', 'AI', 'Data Science', 'python'],
  install_requires=[
    'argparse',
	'ipykernel',
	'ipython',
	'ipython-genutils',
	'numpy',
	'pandas',
	'scikit-learn',
	'fastdtw',
	'dtwalign',
	'tzlocal',
	],
  classifiers=[
    'Development Status :: 3 - Alpha',      # "3 - Alpha", "4 - Beta" or "5 - Production/Stable" -> current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
	'Programming Language :: Python :: 3.7',
	'Programming Language :: Python :: 3.8',
	'Programming Language :: Python :: 3.9',
	
  ],
  package_dir={"": "src"},
  packages=setuptools.find_packages(where="src"),
)


import pandas as pd

from ..filters.Filter import Filter


class FillTimestampsFilter(Filter):

     def applyFilter(self, dataFrame: pd.DataFrame):
        
        if dataFrame.index.isna().any():
            tempIndex = dataFrame.index.to_series()
            time = dataFrame.index.to_series().copy()
            diff = time.diff().median()
            for x in range(1, time.size):
                time.iloc[x] = time.iloc[x-1] + diff

            tempIndex.fillna(value=time, inplace=True)
        else:
            print("[fillTimestamps-Filter]: There are no NaT's to fill!")

        return dataFrame



import pandas as pd
import numpy as np
from sklearn.neighbors import KNeighborsRegressor

from .Filter import Filter

class NoiseFilter(Filter):
    def applyFilter(self, dataFrame:pd.DataFrame, columnlist):
        # check if columnlist is empty (all columns selected) or if certain columns are selected
        # if empty --> ALL dataframe columns are assigned to columnlist
        if len(columnlist) == 0:
            columnlist = dataFrame.columns
        else:
            pass

        # iterate through columnlist 
        for column in columnlist:
            # use a KNN ML model to reconstruct data and denoising it
            clf = KNeighborsRegressor(n_neighbors=100, weights='uniform')
            values_list = dataFrame.iloc[:,column].to_list()
            values_list = list(np.float_(values_list))
            # dataFrame.index.dropna(inplace = True)
            clf.fit(dataFrame.index.values[:, np.newaxis], values_list)
            column_new = clf.predict(dataFrame.index.values[:, np.newaxis])
            # old column is replaced by new column
            dataFrame.iloc[:,column] = column_new

        return dataFrame
    